from aloft.chart import get_charts_to_use
from aloft.cluster import get_default_domain_from_cluster_config


def get_apply_chart_args(arguments):
    # TODO: get volume lock option
    return get_chart_args(arguments)


def get_delete_chart_args(arguments):
    return get_chart_args(arguments)


def get_chart_args(arguments):
    release_id = arguments['<release_id>']
    chart_set = arguments['<chart_set>']
    sandbox_name = arguments['--sandbox']
    charts = get_charts(arguments, chart_set)
    return release_id, chart_set, charts, sandbox_name


def get_charts(arguments, chart_set):
    charts = arguments['--charts']
    if charts is not None:
        charts = charts.split(',')
    else:
        charts = get_charts_to_use(chart_set)
    return charts


def get_get_current_cluster_args(arguments):
    output_type = arguments['--output']
    return output_type


def get_get_clusters_args(arguments):
    domain = arguments['<domain>']
    output_type = arguments['--output']
    if domain is None:
        domain = get_default_domain_from_cluster_config()
    return domain, output_type


def get_create_cluster_args(arguments):
    return get_cluster_args(arguments)


def get_use_cluster_args(arguments):
    return get_cluster_args(arguments)


def get_delete_cluster_args(arguments):
    return get_cluster_args(arguments)


def get_cluster_args(arguments):
    cluster_id = arguments['<cluster_id>']
    return cluster_id



import os
import yaml
from aloft.aws import set_aws_profile
from aloft.input import get_directory_from_environment
from aloft.output import print_table, print_list_as_yaml
from aloft.process import execute


def get_and_print_current_cluster(output_type):
    current_cluster_id = get_current_cluster_id()
    cluster = get_cluster_by_cluster_id(current_cluster_id)
    print_clusters([cluster], output_type)


def get_current_cluster_id():
    output = execute('kubectl config current-context', quiet=True)
    return output.rstrip()


def get_and_print_clusters(cluster_domain, output_type):
    clusters = get_clusters(cluster_domain)
    print_clusters(clusters, output_type)


def print_clusters(clusters, output_type):
    if output_type == 'yaml':
        print_list_as_yaml(clusters)
    elif output_type == 'name':
        print_cluster_names(clusters)
    else:
        print_cluster_text(clusters)


def print_cluster_names(clusters):
    for cluster in clusters:
        print(cluster['metadata']['name'])


def print_cluster_text(clusters):
    rows = []

    for cluster in clusters:
        rows.append([
            cluster['metadata']['name'],
            cluster['spec']['kubernetesVersion'],
            cluster['spec']['networkCIDR'],
        ])

    print_table(['NAME', 'VERSION', 'CIDR'], rows, '{0: <40} {1: <10} {2: <18}')


def get_cluster_by_cluster_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    clusters = get_clusters_by_vpc_id(vpc_id)
    matching_clusters = list(filter(lambda cluster: cluster['metadata']['name'] == cluster_id, clusters))

    if len(matching_clusters) != 1:
        raise KeyError(f'Unable to get cluster data for: {cluster_id}')

    return matching_clusters[0]


def get_clusters(cluster_domain):
    vpc_names = get_vpc_names(cluster_domain)
    clusters = []

    for vpc_name in vpc_names:
        vpc_id = get_vpc_id_by_vpc_name_and_cluster_domain(vpc_name, cluster_domain)
        clusters += get_clusters_by_vpc_id(vpc_id)

    return clusters


def get_clusters_by_vpc_id(vpc_id):
    set_aws_profile_by_vpc_id(vpc_id)
    state_store_locator = get_state_store_locator(vpc_id)
    output = execute(f'kops get clusters -o yaml --state {state_store_locator}', ['no clusters found'], quiet=True)
    clusters = []

    if output:
        for cluster_definition in output.split('---'):
            cluster = yaml.load(cluster_definition)
            clusters.append(cluster)

    return clusters


def     create_cluster(cluster_id):
    print(f'created cluster {cluster_id}')


def use_cluster(cluster_id):
    set_aws_profile_by_cluster_id(cluster_id)
    state_store_locator = get_state_store_locator_by_cluster_id(cluster_id)
    execute(f'kops export kubecfg {cluster_id} --state {state_store_locator}')


def delete_cluster(cluster_id):
    set_aws_profile_by_cluster_id(cluster_id)
    state_store_locator = get_state_store_locator_by_cluster_id(cluster_id)
    execute(f'kops delete cluster {cluster_id} --yes --state {state_store_locator}')


def create_namespace(namespace):
    execute(f'kubectl create namespace {namespace}', ['AlreadyExists'])


def delete_namespace(namespace):
    execute(f'kubectl delete namespace {namespace}', ['NotFound'])


def get_state_store_locator_by_cluster_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    return get_state_store_locator(vpc_id)


def get_state_store_locator(vpc_id):
    return f's3://k8s-state.{vpc_id}'


def set_aws_profile_by_cluster_id(cluster_id):
    aws_profile_name = get_aws_profile_name_by_cluster_id(cluster_id)
    set_aws_profile(aws_profile_name)


def get_aws_profile_name_by_cluster_id(cluster_id):
    return get_vpc_name_by_custer_id(cluster_id)


def set_aws_profile_by_vpc_id(vpc_id):
    aws_profile_name = get_aws_profile_name_by_vpc_id(vpc_id)
    set_aws_profile(aws_profile_name)


def get_aws_profile_name_by_vpc_id(vpc_id):
    return get_vpc_name_by_vpc_id(vpc_id)


def get_vpc_id_by_vpc_name_and_cluster_domain(vpc_name, cluster_domain):
    return f'{vpc_name}.{cluster_domain}'


def get_vpc_name_by_custer_id(cluster_id):
    vpc_id = get_vpc_id_by_cluster_id(cluster_id)
    return get_vpc_name_by_vpc_id(vpc_id)


def get_vpc_name_by_vpc_id(vpc_id):
    return vpc_id.split('.', 1)[0]


def get_vpc_id_by_cluster_id(cluster_id):
    return cluster_id.split('.', 1)[1]


def get_vpc_names(domain):
    domain_cluster_config_directory = get_domain_cluster_config_directory(domain)
    vpc_names = os.listdir(domain_cluster_config_directory)

    if len(vpc_names) == 0:
        raise ValueError(f'No cluster configuration found in {domain_cluster_config_directory}.')

    vpc_names.sort()

    return vpc_names


def get_domain(cluster_id):
    return cluster_id.split('.')[-2:]


def get_base_cluster_config_directory():
    return get_directory_from_environment('K8S_CLUSTER_CONFIG') + '/config'


def get_domain_cluster_config_directory(domain):
    base_cluster_config_directory = get_base_cluster_config_directory()
    return f'{base_cluster_config_directory}/{domain}'


def get_default_domain_from_cluster_config():
    base_cluster_config_directory = get_base_cluster_config_directory()
    domains = os.listdir(base_cluster_config_directory)

    if len(domains) > 1:
        raise ValueError(f'More than one domain config directory found in {base_cluster_config_directory}. You must '
                         f'specify a <domain>')
    elif not domains:
        raise ValueError(f'No cluster domain configurations found in {base_cluster_config_directory}.')

    return domains[0]



import os
import yaml
from io import open
from tempfile import NamedTemporaryFile
from aloft.process import execute
from aloft.cluster import use_cluster, create_namespace, delete_namespace
from aloft.input import get_directory_from_environment, render_template_file


def get_charts_to_use(chart_set):
    charts_config = get_charts_config(chart_set)
    charts_to_apply = []

    for chart in charts_config['charts']:
        charts_to_apply.append(chart['name'])

    return charts_to_apply


def get_charts_config(chart_set):
    chart_set_config_directory = get_chart_set_config_directory(chart_set)
    with open(f'{chart_set_config_directory}/charts.yaml', 'r') as stream:
        return yaml.load(stream)


def get_chart_set_config_directory(chart_set):
    base_chart_config_directory = get_base_chart_config_directory()
    return f'{base_chart_config_directory}/chart-sets/{chart_set}'


def get_base_chart_config_directory():
    return get_directory_from_environment('K8S_CHART_CONFIG')


def apply_charts(release_id, chart_set, charts, sandbox_name):
    release_config = get_release_config(chart_set, release_id, sandbox_name)
    namespace = release_config['namespace']
    use_cluster(release_config['cluster'])
    create_namespace(namespace)

    for chart in charts:
        # TODO: Apply any persistent volume config files
        # kubectl apply -f "$releaseConfigsDirectory/"

        value_filenames = generate_value_files(release_id, chart_set, chart, release_config)
        apply_chart(chart, namespace, value_filenames)

        # TODO: Lock volumes if option is set
        # if [ "$lockVolumes" = "true" ]; then
        # ./lock-volumes $releaseId $chartSetName $chart
        # fi


def add_sandbox_properties(release_config, sandbox_name):
    if sandbox_name:
        sandbox_name_prefixed_with_dash = f'-{sandbox_name}'
        sandbox_name_suffixed_with_dash = f'{sandbox_name}-'
    else:
        sandbox_name = ''
        sandbox_name_prefixed_with_dash = ''
        sandbox_name_suffixed_with_dash = ''

    release_config['properties']['sandboxName'] = sandbox_name
    release_config['properties']['sandboxNamePrefixedWithDash'] = sandbox_name_prefixed_with_dash
    release_config['properties']['sandboxNameSuffixedWithDash'] = sandbox_name_suffixed_with_dash


def get_release_config(chart_set, release_id, sandbox_name):
    release_configs = get_releases_config(chart_set)
    release_config = release_configs[release_id]

    if sandbox_name:
        sandbox_config = release_configs['sandbox']
        sandbox_config['valuesPath'] = release_config['valuesPath']
        sandbox_config['namespace'] += f'-{sandbox_name}'
        release_config = sandbox_config

    add_sandbox_properties(release_config, sandbox_name)

    return release_config


def get_releases_config(chart_set):
    chart_set_config_directory = get_chart_set_config_directory(chart_set)
    with open(f'{chart_set_config_directory}/releases.yaml', 'r') as stream:
        return yaml.load(stream)


def generate_value_files(release_id, chart_set, chart, release_config):
    values_config_directory = get_chart_set_config_directory(chart_set)
    value_paths = release_config.get('valuesPath', f'values/{release_id}').split('/')
    properties = release_config['properties']
    value_filenames = []

    for path in value_paths:
        values_config_directory += f'/{path}'
        append_generated_values_file(values_config_directory, chart, properties, value_filenames)

    return value_filenames


def append_generated_values_file(values_config_directory, chart, properties, value_filenames):
    filename = f'{values_config_directory}/{chart}.yaml'

    if os.path.isfile(filename):
        rendered_values = render_template_file(filename, properties)
        with NamedTemporaryFile(mode='w+t', encoding='utf-8', delete=False) as tmp:
            tmp.write(rendered_values)
            tmp.write('\n')
            value_filenames.append(tmp.name)


def apply_chart(chart, namespace, value_filenames):
    base_chart_config_directory = get_base_chart_config_directory()
    chart_directory = f'{base_chart_config_directory}/charts/{chart}'
    value_args = ' '.join(map(lambda filename: f'-f {filename}', value_filenames))

    execute(f'helm dependencies build {chart_directory}')
    execute(f'helm upgrade --wait -i {namespace}-{chart} --namespace {namespace} {chart_directory} {value_args}')


def delete_charts(release_id, chart_set, charts, sandbox_name):
    release_config = get_release_config(chart_set, release_id, sandbox_name)
    use_cluster(release_config['cluster'])

    for chart in charts:
        delete_chart(chart, release_config['namespace'])

    delete_namespace_if_empty(release_config['namespace'])


def delete_chart(chart, namespace):
    execute(f'helm delete --purge {namespace}-{chart}', ['not found'])

    # TODO: delete PV if volume lock config file exists
    # volumeLockFile="$releaseConfigsDirectory/$chart-pv.yaml"
    #
    # if [ -f $volumeLockFile ]; then
    #     kubectl delete -f $volumeLockFile
    # fi


def delete_namespace_if_empty(namespace):
    releases = get_releases_by_namespace(namespace)

    if not releases:
        delete_namespace(namespace)


def get_releases_by_namespace(namespace):
    output = execute(f'helm ls -q --namespace {namespace}')
    return output.splitlines()



from subprocess import Popen, PIPE, CalledProcessError


def execute(cmd, ignore_error_messages=None, quiet=False):
    process = Popen(cmd, stderr=PIPE, stdout=PIPE, shell=True, encoding="utf-8")
    stderr, stdout = communicate(process, cmd, quiet)
    return_code = validate_return_code(process.returncode, ignore_error_messages, stderr, quiet)
    print_stderr(return_code, stderr, quiet)
    raise_if_error(return_code, cmd, stdout, stderr)

    return stdout


def communicate(process, cmd, quiet):
    stdout = ""

    if not quiet:
        print(f'\n**** {cmd}\n')

    for line in process.stdout:
        print_output(line.rstrip(), quiet)
        stdout += line

    stdout_empty, stderr = process.communicate()

    return stderr, stdout


def print_stderr(return_code, stderr, quiet):
    if return_code != 0:
        quiet = False

    if stderr:
        if not quiet:
            print("\nERROR:")
        for line in iter(stderr.splitlines()):
            print_output(line, quiet)
        print_output('', quiet)


def print_output(line, quiet):
    if not quiet:
        print(f'     {line}')


def validate_return_code(return_code, ignore_error_messages, stderr, quiet):
    if return_code != 0 and ignore_error_messages:
        for ignore_error_message in ignore_error_messages:
            if ignore_error_message in stderr:
                if not quiet:
                    print(f'\n** Ignoring return code: {return_code} - {ignore_error_messages[0]}')
                return_code = 0
                break

    return return_code


def raise_if_error(return_code, cmd, stdout, stderr):
    if return_code != 0:
        raise CalledProcessError(return_code, cmd, stdout, stderr)



import jinja2
import os


def get_directory_from_environment(environment_key):
    base_cluster_config_directory = os.environ[environment_key]
    validate_config_directory(base_cluster_config_directory, environment_key)
    return base_cluster_config_directory


def validate_config_directory(config_directory, environment_key):
    if config_directory is None:
        raise NotADirectoryError(f'{environment_key} environment variable has not been set')
    elif not os.path.isdir(config_directory):
        raise NotADirectoryError(f'{environment_key} must reference a directory: {config_directory}')


def render_template_file(template_filename, context):
    path, filename = os.path.split(template_filename)
    return jinja2.Environment(
        loader=jinja2.FileSystemLoader(path)
    ).get_template(filename).render(context)



import yaml


def print_table(headings, rows, row_format):
    print(row_format.format(*headings))

    for row in rows:
        print(row_format.format(*row))


def print_list_as_yaml(output_list):
    print('\n---\n'.join(map(dump_yaml, output_list)))


def dump_yaml(object_to_dump):
    output = yaml.dump(object_to_dump)
    return output.rstrip()



from setuptools import setup

setup(name='aloft',
      version='0.1.0',
      description='Tool to manage Kubernetes clusters and helm charts across multiple AWS accounts and clusters',
      packages=['aloft'],
      scripts=['bin/aloft'],
      zip_safe=False)
